/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id : $ 
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.se.bonita;

import hero.client.test.SimpleCallbackHandler;
import hero.interfaces.BnNodePropertyValue;
import hero.interfaces.BnNodeValue;
import hero.interfaces.BnProjectLightValue;
import hero.interfaces.BnProjectPropertyValue;
import hero.interfaces.Constants;
import hero.interfaces.ProjectSession;
import hero.interfaces.ProjectSessionHome;
import hero.interfaces.UserSession;
import hero.interfaces.UserSessionHome;
import hero.util.HeroException;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;
import java.util.Properties;

import javax.ejb.CreateException;
import javax.jbi.JBIException;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.rmi.PortableRemoteObject;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;

import org.objectweb.petals.common.descriptor.Extensions;
import org.objectweb.petals.common.util.StringHelper;
import org.objectweb.petals.common.util.XMLUtil;
import org.objectweb.petals.component.common.HandlingException;
import org.objectweb.petals.component.common.se.AbstractServiceEngine;
import org.objectweb.petals.component.common.util.MessageExchangeWrapper;
import org.objectweb.petals.component.common.util.PetalsExtensionsUtil;
import org.objectweb.security.context.SecurityContext;
import org.objectweb.security.context.SecurityCurrent;
import org.ow2.petals.se.bonita.security.JavaSecurityUtil;
import org.w3c.dom.DOMException;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * <h1>Bonita Engine</h1>
 * 
 * 
 * <h1>1. Purpose</h1>
 * 
 * <p>
 * This service engine is used to interact with Bonita workflows: 
 * list projects, list activities, instanciate projects, start and stop activities, set workflow properties.
 * See <a href="http://wiki.bonita.objectweb.org">http://wiki.bonita.objectweb.org</a>.
 * </p>
 * 
 * 
 * <h1>2. Operations</h1>
 * 
 * 
 * <h2>listProjects</h2>
 * 
 * <h3>Purpose</h3>
 * <p>
 * Return a list of projects (Bonita process models). Optional: include the initial properties
 * (returnProperties = true) and list only the models startable by the user 
 * (startableOnly = true for using the initiator mappers).
 * </p>
 * 
 * <h3>Message format</h3>
 * <pre>&lt;content&gt;
    &lt;login&gt;bsoa&lt;/login&gt;
    &lt;returnProperties&gt;true&lt;/returnProperties&gt;
    &lt;startableOnly&gt;true&lt;/startableOnly&gt;
&lt;/content&gt;</pre>
 * 
 * <h3>Engine return</h3>
 * <pre>&lt;content&gt;
    &lt;process projectName="..."&gt;
        &lt;key1&gt;value&lt;/key1&gt;
        &lt;key2&gt;value&lt;/key2&gt;
    &lt;/process&gt;
    &lt;process projectName="..."&gt;
        &lt;key1&gt;value&lt;/key1&gt;
        &lt;key2&gt;value&lt;/key2&gt;
    &lt;/process&gt;
    &lt;process projectName="..."&gt;
        &lt;key1&gt;value&lt;/key1&gt;
        &lt;key2&gt;value&lt;/key2&gt;
    &lt;/process&gt;
&lt;/content&gt;</pre>
 * 
 * 
 * <h2>getProjectProperties</h2>
 * 
 * <h3>Purpose</h3>
 * <p>
 * Return the initial project properties.
 * </p>
 * 
 * <h3>Message format</h3>
 * <pre>&lt;content&gt;
    &lt;process projectName="MyBonitaWorkflow"/&gt;
&lt;/content&gt;</pre>
 * 
 * <h3>Engine return</h3>
 * <pre>&lt;content&gt;
    &lt;process projectName="MyBonitaWorkflow"&gt;
        &lt;key1&gt;value&lt;/key1&gt;
        &lt;key2&gt;value&lt;/key2&gt;
    &lt;/process&gt;
&lt;/content&gt;</pre>
 * 
 * 
 * <h2>instanciateProject</h2>
 * 
 * <h3>Purpose</h3>
 * <p>
 * This operation creates a new instance of the specified project. The instance properties
 * and the activity properties (optional) are set with transmitted values.
 * </p>
 * 
 * <h3>Message format - sample 1</h3>
 * <pre>&lt;content&gt;
    &lt;process projectName="MyBonitaWorkflow"&gt;
        &lt;key1&gt;value&lt;/key1&gt;
    &lt;/process&gt;
&lt;/content&gt;</pre>
 * 
 * <h3>Message format - sample 2</h3>
 * <pre>&lt;content&gt;
    &lt;process projectName="MyBonitaWorkflow" projectVersion="1.1"&gt;
        &lt;key1&gt;value&lt;/key1&gt;
    &lt;/process&gt;
    &lt;activity activityName="FirstActivity"&gt;
        &lt;key3&gt;value&lt;/key3&gt;
    &lt;/activity&gt;
&lt;/content&gt;</pre>
 * 
 * <h3>Engine return</h3>
 * <p>The engine returns the instance name and all properties:</p>
 * <pre>&lt;content&gt;
    &lt;process projectName="MyBonitaWorkflow" instanceName="MyBonitaWorkflow_instance36"&gt;
        &lt;key1&gt;value&lt;/key1&gt;
        &lt;key2&gt;defaultValue&lt;/key2&gt;
    &lt;/process&gt;
    &lt;activity activityName="FirstActivity" projectName="MyBonitaWorkflow" instanceName="MyBonitaWorkflow_instance36"&gt;
        &lt;key3&gt;value&lt;/key3&gt;
    &lt;/activity&gt;
&lt;/content&gt;</pre>
 * 
 * 
 * <h2>listUserActivities</h2>
 * 
 * <h3>Purpose</h3>
 * <p>
 * Return a list of activities of a user (to do list). Optional: include the activity properties
 * (returnProperties = true) and include terminated activities 
 * (returnTerminatedActivities = true).
 * </p>
 * 
 * <h3>Message format</h3>
 * <pre>&lt;content&gt;
    &lt;login&gt;bsoa&lt;/login&gt;
    &lt;returnProperties&gt;true&lt;/returnProperties&gt;
    &lt;returnTerminatedActivities&gt;true&lt;/returnTerminatedActivities&gt;
&lt;/content&gt;</pre>
 * 
 * <h3>Engine return</h3>
 * <pre>&lt;content&gt;
    &lt;activity activityName="..." projectName="..." instanceName="..." state="EXECUTING"&gt;
        &lt;key1&gt;value&lt;/key1&gt;
        &lt;key2&gt;value&lt;/key2&gt;
    &lt;/activity&gt;

    &lt;activity activityName="..." projectName="..." instanceName="..." state="TERMINATED"&gt;
        &lt;key1&gt;value&lt;/key1&gt;
        &lt;key2&gt;value&lt;/key2&gt;
    &lt;/activity&gt;
&lt;/content&gt;</pre>
 * 
 * 
 * <h2>startActivity</h2>
 * 
 * <h3>Purpose</h3>
 * <p>
 * Bonita Engine can start the activities who have the state "Ready". The instance properties
 * and the activity properties are updated with transmitted values.
 * </p>
 * 
 * <h3>Message format</h3>
 * <pre>&lt;content&gt;
    &lt;process instanceName="MyBonitaWorkflow_instance36/"&gt;
    &lt;activity activityName="MyActivity"&gt;
        &lt;key2&gt;value&lt;/key2&gt;
    &lt;/activity&gt;
&lt;/content&gt;</pre>
 *  
 * <h3>Engine return</h3>
 * <pre>&lt;content&gt;
    &lt;process projectName="MyBonitaWorkflow" instanceName="MyBonitaWorkflow_instance36"&gt;
        &lt;key1&gt;defaultValue&lt;/key1&gt;
    &lt;/process&gt;
    &lt;activity activityName="MyActivity" projectName="MyBonitaWorkflow" instanceName="MyBonitaWorkflow_instance36"&gt;
        &lt;key2&gt;value&lt;/key2&gt;
        &lt;key3&gt;defaultValue&lt;/key3&gt;
    &lt;/activity&gt;
&lt;/content&gt;</pre>
 * 
 * 
 * <h2>getActivityProperties</h2>
 * 
 * <h3>Purpose</h3>
 * <p>
 * Return the instance properties and the activity properties.
 * </p>
 * 
 * <h3>Message format</h3>
 * <pre>&lt;content&gt;
    &lt;process instanceName="MyBonitaWorkflow_instance36"/&gt;
    &lt;activity activityName="MyActivity"/&gt;
&lt;/content&gt;</pre>
 * 
 * <h3>Engine return</h3>
 * <pre>&lt;content&gt;
    &lt;process projectName="MyBonitaWorkflow" instanceName="MyBonitaWorkflow_instance36"&gt;
        &lt;key1&gt;value&lt;/key1&gt;
        &lt;key2&gt;value&lt;/key2&gt;
    &lt;/process&gt;
    &lt;activity activityName="MyActivity" projectName="MyBonitaWorkflow" instanceName="MyBonitaWorkflow_instance36"&gt;
        &lt;key3&gt;value&lt;/key3&gt;
        &lt;key4&gt;value&lt;/key4&gt;
    &lt;/activity&gt;
&lt;/content&gt;</pre>
 * 
 * 
 * <h2>setActivityProperties</h2>
 * 
 * <h3>Purpose</h3>
 * <p>
 * Set the activity properties and the instance properties with the specified values.
 * </p>
 * 
 * <h3>Message format</h3>
 * <pre>&lt;content&gt;
    &lt;process instanceName="MyBonitaWorkflow_instance36"&gt;
        &lt;key1&gt;value1&lt;/key1&gt;
        &lt;key2&gt;value2&lt;/key2&gt;
    &lt;/process&gt;
    &lt;activity activityName="MyActivity"&gt;
        &lt;key1&gt;11&lt;/key1&gt;
        &lt;key2&gt;22&lt;/key2&gt;
    &lt;/activity&gt;
&lt;/content&gt;</pre>
 * 
 * <h3>Engine return</h3>
 * <p>If there is no exception, the returned message is:</p>
 * <pre>&lt;success/&gt;</pre>
 * 
 * 
 * <h2>terminateActivity</h2>
 * 
 * <h3>Purpose</h3>
 * <p>
 * Bonita Engine can terminate the activities who have the state "Executing". The instance properties
 * and the activity properties are updated with transmitted values.
 * </p>
 * 
 * <h3>Message format</h3>
 * <pre>&lt;content&gt;
    &lt;process instanceName="MyBonitaWorkflow_instance36"/&gt;
    &lt;activity activityName="MyActivity"&gt;
        &lt;key1&gt;value&lt;/key1&gt;
    &lt;/activity&gt;
&lt;/content&gt;</pre>
 * 
 * <h3>Engine return</h3>
 * <p>If there is no exception, the returned message is:</p>
 * <pre>&lt;success/&gt;</pre>
 * 
 * 
 * <h1>3. Service Unit configuration</h1>
 * 
 * <ul>
 * <li><code>rmi-url</code>: URL of the JOnAS RMI registry used to access EJB.</li>
 * <li><code>loginContextName</code>: the login contexts are in the file <code>jaas.config</code></li>
 * <li><code>login</code>: Bonita login of the default user</li>
 * <li><code>password</code>: Bonita password of the default user</li>
 * <li><code>role</code>: Bonita role of the default user (ex.: BONITAUSER)</li>
 * <li><code>loginX</code>: Bonita login of the user X (replace X by a string id)</li>
 * <li><code>passwordX</code>: Bonita password of the user X (replace X by a string id)</li>
 * <li><code>roleX</code>: Bonita role of the user X (replace X by a string id)</li>
 * </ul>
 * 
 * <p>Several users can be specified using different logins and/or passwords (see samples below).
 * There are two authentication methods: direct authentication (with password)
 * or using Jonas propagation context (with role).</p>
 * 
 * <p>Sample of <code>jbi.xml</code> file with login/password authentication:</p>
 * 
 * <pre>&lt;jbi version="1.0" xmlns='http://java.sun.com/xml/ns/jbi'
        xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance'
        xmlns:petals="http://petals.objectweb.org/"
        xmlns:extensions="http://petals.objectweb.org/extensions/"
        xmlns:keyvalue="http://petals.objectweb.org/extensions/key-value/"&gt;
    &lt;services binding-component="false"&gt;
        &lt;provides interface-name="SampleBonita"
                service-name="petals:SampleBonitaService"
                endpoint-name="SampleBonitaEndpoint"&gt;
            &lt;extensions:extensions&gt;
                &lt;keyvalue:extension&gt;
                    &lt;rmi-url&gt;rmi://127.0.0.1:1099&lt;/rmi-url&gt;
                    &lt;loginContextName&gt;bonita&lt;/loginContextName&gt;
                    
                    &lt;login&gt;bsoa&lt;/login&gt;
                    &lt;password&gt;bsoa&lt;/password&gt;
                    
                    &lt;login2&gt;admin&lt;/login2&gt;
                    &lt;password2&gt;admin&lt;/password2&gt;
                    
                    &lt;wsdl&gt;samplebonita.wsdl&lt;/wsdl&gt;
                &lt;/keyvalue:extension&gt;
            &lt;/extensions:extensions&gt;
        &lt;/provides&gt;
    &lt;/services&gt;
&lt;/jbi&gt;</pre>
 * 
 * <p>Sample of <code>jbi.xml</code> file with login/role authentication:</p>
 * 
 * <pre>&lt;jbi version="1.0" xmlns='http://java.sun.com/xml/ns/jbi'
        xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance'
        xmlns:petals="http://petals.objectweb.org/"
        xmlns:extensions="http://petals.objectweb.org/extensions/"
        xmlns:keyvalue="http://petals.objectweb.org/extensions/key-value/"&gt;
    &lt;services binding-component="false"&gt;
        &lt;provides interface-name="SampleBonita"
                service-name="petals:SampleBonitaService"
                endpoint-name="SampleBonitaEndpoint"&gt;
            &lt;extensions:extensions&gt;
                &lt;keyvalue:extension&gt;
                    &lt;rmi-url&gt;rmi://127.0.0.1:1099&lt;/rmi-url&gt;
                    &lt;loginContextName&gt;bonita-petals&lt;/loginContextName&gt;
                    
                    &lt;login&gt;bsoa&lt;/login&gt;
                    &lt;role&gt;BONITAUSER&lt;/role&gt;
                    
                    &lt;login2&gt;admin&lt;/login2&gt;
                    &lt;role2&gt;BONITAUSER&lt;/role2&gt;
                    
                    &lt;wsdl&gt;samplebonita.wsdl&lt;/wsdl&gt;
                &lt;/keyvalue:extension&gt;
            &lt;/extensions:extensions&gt;
        &lt;/provides&gt;
    &lt;/services&gt;
&lt;/jbi&gt;</pre>
 *
 * <h1>4. Security</h1>
 * 
 * <p>In order to access Bonita, the engine needs two security files:</p>
 * 
 * <ul>
 *     <li>{user.home}/petals-bonita-engine_default_java.login.config</li>
 *     <li>{user.home}/petals-bonita-engine_default_java.policy</li>
 * </ul>
 * 
 * <p>If these files do not exist, the engine copies in the user home
 * the default files from the engine jar. If you want to customize these
 * files, put your customized files in the user home, with the same
 * filenames as above.</p>
 * 
 * <p>For the propagation context (login/role authentication), you have to add
 * a login context in the file $JONAS_BASE/conf/jaas.config:</p>
 * 
 * <pre>bonita-petals {
    org.ow2.petals.se.bonita.security.PetalsSeBonitaDefaultLoginModule required
	;
    // Use the login module to propagate security to the JOnAS server
    org.objectweb.jonas.security.auth.spi.ClientLoginModule  required
	;
};</pre>
 * 
 * 
 * <p>The propagation must be activated in the file $JONAS_BASE/conf/jonas.properties:</p>
 * 
 * <pre>jonas.security.propagation              true
jonas.transaction.propagation           true</pre>
 * 
 * @version $Rev: $ $Date: $
 * @author Guillaume Decarnin - Open Wide
 *
 */
public class BonitaEngine extends AbstractServiceEngine {
    
    // operations
    private static final String LIST_PROJECTS_OPERATION = "listProjects";
    private static final String GET_PROJECT_PROPERTIES_OPERATION = "getProjectProperties";
    private static final String INSTANCIATE_PROJECT_OPERATION = "instanciateProject";
    private static final String LIST_USER_ACTIVITIES_OPERATION = "listUserActivities";
    private static final String START_ACTIVITY_OPERATION = "startActivity";
    private static final String GET_ACTIVITY_PROPERTIES_OPERATION = "getActivityProperties";
    private static final String SET_ACTIVITY_PROPERTIES_OPERATION = "setActivityProperties";
    private static final String TERMINATE_ACTIVITY_OPERATION = "terminateActivity";
    
    // keys for key-value extension
    public static final String LOGIN_KEY = "login";
    public static final String PASSWORD_KEY = "password";
    public static final String ROLE_KEY = "role";
    public static final String RMI_URL_KEY = "rmi-url";
    public static final String LOGIN_CONTEXT_NAME_KEY = "loginContextName";

    private static final JavaSecurityUtil bonitaEngineJavaSecurityUtil = new JavaSecurityUtil("petals-bonita-engine_default_");
    
    @Override
    protected void init() throws JBIException {
        setServiceUnitExtensionsValidator(new BonitaServiceUnitExtensionsValidator());
    }
    
    @Override
    protected boolean onExchange(MessageExchangeWrapper exchange, Extensions extensions) throws HandlingException {
        String operation = exchange.getOperationName();

        try {
            Node message = BonitaXmlUtil.createDOMNodeFromSource(exchange.getInMessage().getContent()).getFirstChild();
            
            if (operation.equals(LIST_PROJECTS_OPERATION)) {
                return doListProjects(message, exchange, extensions);
            } else if (operation.equals(GET_PROJECT_PROPERTIES_OPERATION)) {
                return doGetProjectProperties(message, exchange, extensions);
            } else if (operation.equals(INSTANCIATE_PROJECT_OPERATION)) {
                return doInstanciateProject(message, exchange, extensions);
            } else if (operation.equals(LIST_USER_ACTIVITIES_OPERATION)) {
                return doListUserActivities(message, exchange, extensions);
            } else if (operation.equals(START_ACTIVITY_OPERATION)) {
                return doStartActivity(message, exchange, extensions);
            } else if (operation.equals(GET_ACTIVITY_PROPERTIES_OPERATION)) {
                return doGetActivityProperties(message, exchange, extensions);
            } else if (operation.equals(SET_ACTIVITY_PROPERTIES_OPERATION)) {
                return doSetActivityProperties(message, exchange, extensions);
            } else if (operation.equals(TERMINATE_ACTIVITY_OPERATION)) {
                return doTerminateActivity(message, exchange, extensions);
            } else {
                throw new HandlingException("Unknown operation: \"" + operation + "\".");
            }
        } catch (HandlingException e) {
            throw e;
        } catch (Exception e) {
            throw new HandlingException(e);
        }
    }

    private InitialContext getInitialContextAndLogin(Extensions extensions, String login) throws HandlingException, NamingException, LoginException {
    	// checking Java security (policy, jaas) and if none outputting a default one
    	bonitaEngineJavaSecurityUtil.checkJavaSecurityConfiguration();
    	
        // setting up JNDI
        // TODO put all props in env so it will be taken by jaas login's own InitCtx (TODO doc)
        // but do not override them if they already exit (so it may be set at the whole petals system level)
        Properties props = new Properties();
        props.put(Context.INITIAL_CONTEXT_FACTORY, extensionProperties.getAttributeValue(Context.INITIAL_CONTEXT_FACTORY));
        props.put(Context.URL_PKG_PREFIXES, extensionProperties.getAttributeValue(Context.URL_PKG_PREFIXES));
        props.put(Context.PROVIDER_URL, getRmiUrl(extensions));
        
        if (System.getProperty(Context.INITIAL_CONTEXT_FACTORY) == null) {
            System.setProperty(Context.INITIAL_CONTEXT_FACTORY, extensionProperties.getAttributeValue(Context.INITIAL_CONTEXT_FACTORY));
        }
        if (System.getProperty(Context.URL_PKG_PREFIXES) == null) {
            System.setProperty(Context.URL_PKG_PREFIXES, extensionProperties.getAttributeValue(Context.URL_PKG_PREFIXES));
        }
        if (System.getProperty(Context.PROVIDER_URL) == null) {
            System.setProperty(Context.PROVIDER_URL, getRmiUrl(extensions));
        }
        
        InitialContext ic = new InitialContext(props);
        
        login(extensions, login);
        
        return ic;
    }
    
    private static void login(Extensions extensions, String login) throws LoginException, NamingException, HandlingException {
    	// build the login-password map
    	Map<String, String> keysValues = PetalsExtensionsUtil.getExtensionsKeysValues(extensions);
        Map<String, String> loginRoleMap = new HashMap<String, String>();
        Map<String, String> loginPasswordMap = new HashMap<String, String>();
        for (Map.Entry<String, String> entry : keysValues.entrySet()) {
            if (entry.getKey().startsWith(LOGIN_KEY) && !entry.getKey().equals(LOGIN_CONTEXT_NAME_KEY)) {
                String id = entry.getKey().substring(LOGIN_KEY.length());
                String roleKey = ROLE_KEY + id;
                String passwordKey = PASSWORD_KEY + id;
                
                String theLogin    = entry.getValue();
                String theRole     = PetalsExtensionsUtil.extractValueFromKeyValueExtension(extensions, roleKey);
                String thePassword = PetalsExtensionsUtil.extractValueFromKeyValueExtension(extensions, passwordKey);
                loginRoleMap.put(theLogin, theRole);
                loginPasswordMap.put(theLogin, thePassword);
            }
        }
        
    	// find login
    	if (login == null || login.length() == 0) {
            // default login
            login = PetalsExtensionsUtil.extractValueFromKeyValueExtension(extensions, LOGIN_KEY);
        } else if (!loginRoleMap.containsKey(login)) {
        	throw new LoginException("Unknown login \"" + login + "\".");
        }
    	
    	// find password and/or role of this login
        String role = loginRoleMap.get(login);
        boolean useContextPropagation = role != null && role.trim().length() > 0;
        
        // create CallbackHandler and SecurityContext
        CallbackHandler handler = null;
        if (useContextPropagation) {
            SecurityCurrent current = SecurityCurrent.getCurrent();
            if (current == null) {
            	current = new SecurityCurrent();
            }
            SecurityContext sc = new SecurityContext(login, new String[] { role });
        	current.setGlobalSecurityContext(sc);
        	handler = new SimpleCallbackHandler(login, null);
        } else {
        	String password = loginPasswordMap.get(login);
            handler = new SimpleCallbackHandler(login, password.toCharArray());
        }
        
        // create LoginContext and do login
        String loginContextName = PetalsExtensionsUtil.extractValueFromKeyValueExtension(extensions, LOGIN_CONTEXT_NAME_KEY, "bonita"); // as in file "jaas.config"
        LoginContext lc = new LoginContext(loginContextName, handler);
        lc.login();
    }
    
    private static String getRmiUrl(Extensions extensions) throws HandlingException {
        String rmiUrl = PetalsExtensionsUtil.extractValueFromKeyValueExtension(extensions, RMI_URL_KEY);
        if (rmiUrl == null || rmiUrl.trim().length() == 0) {
            throw new HandlingException("SU configuration error: missing '"+ RMI_URL_KEY + "'.");
        }
        return rmiUrl;
    }
    
    private static ProjectSession getProjectSession(InitialContext ic, Extensions extensions)
            throws NamingException, RemoteException, CreateException, HandlingException {
        Object psHome = ic.lookup(getRmiUrl(extensions) + "/ejb/hero/ProjectSession");
        ProjectSessionHome prjHome = (ProjectSessionHome) PortableRemoteObject.narrow(psHome, ProjectSessionHome.class);
        ProjectSession projectSession = prjHome.create();
        return projectSession;
    }
    
    private static UserSession getUserSession(InitialContext ic, Extensions extensions)
            throws NamingException, RemoteException, CreateException, HandlingException {
        Object usHome = ic.lookup(getRmiUrl(extensions) + "/ejb/hero/UserSession");
        UserSessionHome usrHome = (UserSessionHome) PortableRemoteObject.narrow(usHome, UserSessionHome.class);
        UserSession userSession = usrHome.create();
        return userSession;
    }
    
    @SuppressWarnings("unchecked")
    private static void setProjectProperties(ProjectSession projectSession, BonitaProject bonitaProject)
            throws RemoteException, DOMException, HeroException {
        Map<String, String> map = bonitaProject.getProjectProperties();
        if (map.size() == 0) {
            return;
        }
        
        Collection<BnProjectPropertyValue> properties = projectSession.getProperties();
        
        for (BnProjectPropertyValue property : properties) {
            String key = property.getTheKey();
            if (map.containsKey(key)) {
                // update value in Bonita
                property.setTheValue(map.get(key));
                projectSession.setProperty(property.getTheKey(), property.getTheValue());
            }
        }
    }
    
    @SuppressWarnings("unchecked")
    private static void setActivityProperties(ProjectSession projectSession, UserSession userSession, BonitaProject bonitaProject)
            throws RemoteException, DOMException, HeroException {
        if (bonitaProject.getInstanceName() == null || bonitaProject.getActivityName() == null) {
            return;
        }
        
        Map<String, String> map = bonitaProject.getActivityProperties();
        if (map.size() == 0) {
            return;
        }

        BnNodeValue node = userSession.getNode(bonitaProject.getInstanceName(), bonitaProject.getActivityName());
        Collection<BnNodePropertyValue> properties = node.getBnPropertieCollection();
        
        for (BnNodePropertyValue property : properties) {
            String key = property.getTheKey();
            if (map.containsKey(key)) {
                // update value in Bonita
                property.setTheValue(map.get(key));
                projectSession.setNodeProperty(bonitaProject.getActivityName(), property.getTheKey(), property.getTheValue(), property.getPropagate());
            }
        }
    }

    @SuppressWarnings("unchecked")
    private boolean doGetProjectProperties(Node message, MessageExchangeWrapper exchange, Extensions extensions)
            throws LoginException, NamingException, RemoteException, CreateException, HeroException, HandlingException {
        // load data from message
        BonitaProject bonitaProject = new BonitaProject();
        bonitaProject.loadData(message);
        
        String projectName = bonitaProject.getProjectName();
        String projectVersion = bonitaProject.getProjectVersion();

        // check data
        if (projectName == null || projectName.trim().length() == 0) {
            throw new HandlingException("'projectName' attribute not found or empty");
        }
        
        // get EJB sessions
        InitialContext ic = getInitialContextAndLogin(extensions, bonitaProject.getLogin());
        ProjectSession projectSession = getProjectSession(ic, extensions);
        
        // get properties
        if (projectVersion == null || projectVersion.trim().length() == 0) {
            projectSession.initModel(projectName);
        } else {
            projectSession.initModelWithVersion(projectName, projectVersion);
        }
        bonitaProject.getActivityProperties().clear();
        bonitaProject.loadProjectProperties(projectSession);
        
        // build response
        StringBuilder content = new StringBuilder();
        content.append("<content>\n");
        content.append(bonitaProject.processToXml(false));
        content.append("</content>");
        
        // write out message
        exchange.setOutMessageContent(content.toString());
        return true;
    }

    @SuppressWarnings("unchecked")
    private boolean doGetActivityProperties(Node message, MessageExchangeWrapper exchange, Extensions extensions)
            throws LoginException, NamingException, RemoteException, CreateException, HeroException, HandlingException {
        // load data from message
        BonitaProject bonitaProject = new BonitaProject();
        bonitaProject.loadData(message);
        
        String instanceName = bonitaProject.getInstanceName();
        String activityName = bonitaProject.getActivityName();

        // check data
        if (instanceName == null || instanceName.trim().length() == 0) {
            throw new HandlingException("'instanceName' attribute not found or empty");
        }

        if (activityName == null || activityName.trim().length() == 0) {
            throw new HandlingException("'activityName' attribute not found or empty");
        }
        
        // get EJB sessions
        InitialContext ic = getInitialContextAndLogin(extensions, bonitaProject.getLogin());
        ProjectSession projectSession = getProjectSession(ic, extensions);
        
        bonitaProject.loadProjectName(projectSession);
        
        // get properties
        projectSession.initProject(instanceName);
        bonitaProject.loadProjectProperties(projectSession);
        bonitaProject.loadActivityProperties(projectSession);
        bonitaProject.loadActivityState(projectSession);
        
        // write out message
        exchange.setOutMessageContent(bonitaProject.toXml());
        return true;
    }
    
    private boolean doSetActivityProperties(Node message, MessageExchangeWrapper exchange, Extensions extensions)
            throws HandlingException, RemoteException, CreateException, DOMException, HeroException, LoginException, NamingException {
        // load data from message
        BonitaProject bonitaProject = new BonitaProject();
        bonitaProject.loadData(message);
        
        String instanceName = bonitaProject.getInstanceName();
        String activityName = bonitaProject.getActivityName();
        
        // check data
        if (instanceName == null || instanceName.trim().length() == 0) {
            throw new HandlingException("'instanceName' attribute not found or empty");
        }

        if (activityName == null || activityName.trim().length() == 0) {
            throw new HandlingException("'activityName' attribute not found or empty");
        }
        
        // get EJB sessions
        InitialContext ic = getInitialContextAndLogin(extensions, bonitaProject.getLogin());
        ProjectSession projectSession = getProjectSession(ic, extensions);
        UserSession userSession = getUserSession(ic, extensions);
        
        // update Bonita workflow
        projectSession.initProject(instanceName);
        setProjectProperties(projectSession, bonitaProject);
        setActivityProperties(projectSession, userSession, bonitaProject);

        // response
        if (exchange.isInOnlyPattern()) {
            return false;
        }

        // update bonitaProject object
        bonitaProject.loadProjectName(projectSession);
        bonitaProject.loadProjectProperties(projectSession);
        bonitaProject.loadActivityProperties(projectSession);
        bonitaProject.loadActivityState(projectSession);
        
        exchange.setOutMessageContent(bonitaProject.toXml());
        return true;
    }

    @SuppressWarnings("unchecked")
    private boolean doInstanciateProject(Node message, MessageExchangeWrapper exchange, Extensions extensions)
            throws LoginException, NamingException, RemoteException, CreateException, HeroException, HandlingException {
        // load data from message
        BonitaProject bonitaProject = new BonitaProject();
        bonitaProject.loadData(message);

        String projectName = bonitaProject.getProjectName();
        String projectVersion = bonitaProject.getProjectVersion();

        // check data
        if (projectName == null || projectName.trim().length() == 0) {
            throw new HandlingException("'projectName' attribute not found or empty");
        }

        // get EJB sessions
        InitialContext ic = getInitialContextAndLogin(extensions, bonitaProject.getLogin());
        ProjectSession projectSession = getProjectSession(ic, extensions);
        UserSession userSession = getUserSession(ic, extensions);
        
        // update Bonita workflow
        Hashtable map = new Hashtable(bonitaProject.getProjectProperties());
        
        if (projectVersion == null || projectVersion.trim().length() == 0) {
            projectSession.instantiateProject(projectName, map);
        } else {
            projectSession.instantiateProject(projectName, projectVersion, map);
        }
        
        bonitaProject.setInstanceName(projectSession.getName());
        
        setActivityProperties(projectSession, userSession, bonitaProject);

        // response
        if (exchange.isInOnlyPattern()) {
            return false;
        }
        
        // update bonitaProject object
        bonitaProject.setInstanceName(projectSession.getName());
        bonitaProject.loadProjectProperties(projectSession);
        
        if (!StringHelper.isNullOrEmpty(bonitaProject.getActivityName())) {
            bonitaProject.loadActivityProperties(projectSession);
            bonitaProject.loadActivityState(projectSession);
        }
        
        exchange.setOutMessageContent(bonitaProject.toXml());
        return true;
    }
    
    /**
     * On a Ready activity
     * @param message
     * @param exchange
     * @param extensions
     * @throws LoginException
     * @throws NamingException
     * @throws RemoteException
     * @throws CreateException
     * @throws HeroException
     * @throws HandlingException
     */
    private boolean doStartActivity(Node message, MessageExchangeWrapper exchange, Extensions extensions)
            throws LoginException, NamingException, RemoteException, CreateException, HeroException, HandlingException {
        // load data from message
        BonitaProject bonitaProject = new BonitaProject();
        bonitaProject.loadData(message);
        
        String instanceName = bonitaProject.getInstanceName();
        String activityName = bonitaProject.getActivityName();
        
        // check data
        if (instanceName == null || instanceName.trim().length() == 0) {
            throw new HandlingException("'instanceName' attribute not found or empty");
        }
        
        if (activityName == null || activityName.trim().length() == 0) {
            throw new HandlingException("'activityName' attribute not found or empty");
        }
        
        // get EJB sessions
        InitialContext ic = getInitialContextAndLogin(extensions, bonitaProject.getLogin());
        ProjectSession projectSession = getProjectSession(ic, extensions);
        UserSession userSession = getUserSession(ic, extensions);

        // update Bonita workflow
        projectSession.initProject(instanceName);
        userSession.startActivity(instanceName, activityName);
        setProjectProperties(projectSession, bonitaProject);
        setActivityProperties(projectSession, userSession, bonitaProject);

        // response
        if (exchange.isInOnlyPattern()) {
            return false;
        }
        
        // update bonitaProject object
        bonitaProject.loadProjectName(projectSession);
        bonitaProject.loadProjectProperties(projectSession);
        bonitaProject.loadActivityProperties(projectSession);
        bonitaProject.loadActivityState(projectSession);
        
        exchange.setOutMessageContent(bonitaProject.toXml());
        return true;
    }
    
    /**
     * On an Executing activity
     * @param message
     * @param exchange
     * @param extensions
     * @throws LoginException
     * @throws NamingException
     * @throws RemoteException
     * @throws CreateException
     * @throws HeroException
     * @throws HandlingException
     */
    private boolean doTerminateActivity(Node message, MessageExchangeWrapper exchange, Extensions extensions)
            throws LoginException, NamingException, RemoteException, CreateException, HeroException, HandlingException {
        // load data from message
        BonitaProject bonitaProject = new BonitaProject();
        bonitaProject.loadData(message);
        
        String instanceName = bonitaProject.getInstanceName();
        String activityName = bonitaProject.getActivityName();
        
        // check data
        if (instanceName == null || instanceName.trim().length() == 0) {
            throw new HandlingException("'instanceName' attribute not found or empty");
        }
        
        if (activityName == null || activityName.trim().length() == 0) {
            throw new HandlingException("'activityName' attribute not found or empty");
        }
        
        // get EJB sessions
        InitialContext ic = getInitialContextAndLogin(extensions, bonitaProject.getLogin());
        ProjectSession projectSession = getProjectSession(ic, extensions);
        UserSession userSession = getUserSession(ic, extensions);
        
        // check state before setProperties
        projectSession.initProject(instanceName);
        if (projectSession.getNodeState(activityName) != Constants.Nd.EXECUTING) {
            throw new HandlingException("The activity to terminate must be in the EXECUTING state. Call the operation " + START_ACTIVITY_OPERATION + " before " + TERMINATE_ACTIVITY_OPERATION + ".");
        }
        
        // update Bonita workflow
        setProjectProperties(projectSession, bonitaProject);
        setActivityProperties(projectSession, userSession, bonitaProject);
        
        if (!exchange.isInOnlyPattern()) {
            // update bonitaProject object
            bonitaProject.loadProjectName(projectSession);
            bonitaProject.loadProjectProperties(projectSession);
            bonitaProject.loadActivityProperties(projectSession);
            bonitaProject.setActivityState(Constants.Nd.nodeStateName[Constants.Nd.TERMINATED]);
        }
        
        // terminate activity
        userSession.terminateActivity(instanceName, activityName); 

        // response
        if (exchange.isInOnlyPattern()) {
            return false;
        }
        
        exchange.setOutMessageContent(bonitaProject.toXml());
        return true;
    }
    
    @SuppressWarnings("unchecked")
    private boolean doListProjects(Node message, MessageExchangeWrapper exchange, Extensions extensions)
            throws HandlingException, LoginException, NamingException, RemoteException, CreateException, HeroException {
        // get params
        String login = null;
        boolean returnProperties = false;
        boolean startableOnly = false;
        BonitaProject bonitaProject = new BonitaProject();
        
        NodeList children = message.getChildNodes();
        for (int i = 0; i < children.getLength(); i++) {
            Node child = children.item(i);
            if (child.getNodeType() != Node.ELEMENT_NODE) {
                continue;
            }

            if (child.getNodeName().equals("login")) {
                login = XMLUtil.getTextContent(child);
            } else if (child.getNodeName().equals("returnProperties")) {
                returnProperties = Boolean.parseBoolean(XMLUtil.getTextContent(child));
            } else if (child.getNodeName().equals("startableOnly")) {
                startableOnly = Boolean.parseBoolean(XMLUtil.getTextContent(child));
            }
        }
        
        // get EJB sessions
        InitialContext ic = getInitialContextAndLogin(extensions, login);
        ProjectSession projectSession = getProjectSession(ic, extensions);
        UserSession userSession = getUserSession(ic, extensions);
        
        // get models
        Collection<BnProjectLightValue> models;
        if (startableOnly) {
            // with initiator mapper
            models = userSession.getModels();
        } else {
            // without initiator mapper            
            models = userSession.getModelsLight();            
        }

        // build response
        StringBuilder activitiesString = new StringBuilder();
        activitiesString.append("<content>\n");
        
        for (BnProjectLightValue model : models) {
            String projectName = model.getName();
            bonitaProject.setProjectName(projectName);
            
            if (returnProperties) {
                projectSession.initModel(projectName);
                bonitaProject.loadProjectProperties(projectSession);
            }
            
            activitiesString.append(bonitaProject.processToXml(false));
        }
        
        activitiesString.append("</content>");

        // write out message
        exchange.setOutMessageContent(activitiesString.toString());
        return true;
    }
    
    @SuppressWarnings("unchecked")
    private boolean doListUserActivities(Node message, MessageExchangeWrapper exchange, Extensions extensions)
            throws HandlingException, LoginException, NamingException, RemoteException, CreateException, HeroException {
        /* message:
         * <content>
         *     <login>manager</login>
         *     <password>manager</password>
         *     <returnProperties>true</returnProperties>
         *     <returnTerminatedActivities>false</returnTerminatedActivities>
         * </content>
         */
        
        // get params
        String login = null;
        boolean returnProperties = false;
        boolean returnTerminatedActivities = false;
        BonitaProject project = new BonitaProject();
        
        NodeList children = message.getChildNodes();
        for (int i = 0; i < children.getLength(); i++) {
            Node child = children.item(i);
            if (child.getNodeType() != Node.ELEMENT_NODE) {
                continue;
            }

            if (child.getNodeName().equals("login")) {
                login = XMLUtil.getTextContent(child);
            } else if (child.getNodeName().equals("returnProperties")) {
                returnProperties = Boolean.parseBoolean(XMLUtil.getTextContent(child));
            } else if (child.getNodeName().equals("returnTerminatedActivities")) {
                returnTerminatedActivities = Boolean.parseBoolean(XMLUtil.getTextContent(child));
            }
        }
        
        // get EJB sessions
        InitialContext ic = getInitialContextAndLogin(extensions, login);
        ProjectSession projectSession = getProjectSession(ic, extensions);
        UserSession userSession = getUserSession(ic, extensions);
        
        // build response
        StringBuilder activitiesString = new StringBuilder();
        activitiesString.append("<content>\n");
        
        Collection<BnNodeValue> activities = new ArrayList<BnNodeValue>();
        activities.addAll(userSession.getActivityListAllInstances()); // executing state
        activities.addAll(userSession.getToDoListAllInstances()); // ready and anticipable state
        for (BnNodeValue activity : activities) {
            project.setActivityName(activity.getName());
            project.setInstanceName(activity.getBnProject().getName());
            project.loadProjectName(projectSession);
            project.loadActivityState(projectSession);
            
            if (returnProperties) {
                Collection<BnNodePropertyValue> activityProperties = activity.getBnPropertieCollection();
                for (BnNodePropertyValue property : activityProperties) {
                    project.getActivityProperties().put(property.getTheKey(), property.getTheValue());
                }
            }
            
            activitiesString.append(project.activityToXml(true, true, true));
        }
        
        if (returnTerminatedActivities) {
            project.getActivityProperties().clear();
            
            Collection<BnNodeValue> terminatedActivities = userSession.getTerminatedListAllInstances();
            for (BnNodeValue terminatedActivity : terminatedActivities) {
                project.setActivityName(terminatedActivity.getName());
                project.setInstanceName(terminatedActivity.getBnProject().getName());
                project.loadProjectName(projectSession);
                project.loadActivityState(projectSession);
                
                Collection<BnNodePropertyValue> activityProperties = terminatedActivity.getBnPropertieCollection();
                for (BnNodePropertyValue property : activityProperties) {
                    project.getActivityProperties().put(property.getTheKey(), property.getTheValue());
                }
                
                activitiesString.append(project.activityToXml(true, true, true));
            }
        }

        activitiesString.append("</content>");

        // write out message
        exchange.setOutMessageContent(activitiesString.toString());
        return true;
    }

}
