/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id : $ 
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.se.bonita.security;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

/**
 * Checks Java security (policy, jaas, security manager)
 * and if none sets and outputs a default one.
 * To use it, .subclass it and provide the protected fields
 */
public class JavaSecurityUtil {
	
	private static final String JAVA_POLICY_PROP = "java.security.policy";
	private static final String JAAS_CONFIG_PROP = "java.security.auth.login.config";
	
	private static boolean securityChecked = false;
	
	protected String javaPolicyName = "java.policy";
	protected String jaasConfigName = "java.login.config";
	protected String configFileHomePrefix = ".";
	protected String configResourcePrefix = "java_security_util_";
	
	protected String userHome = null;
	
	public JavaSecurityUtil() {}
	
	public JavaSecurityUtil(String configResourcePrefix) {
		this.configResourcePrefix = configResourcePrefix;
	}
	
	protected String getUserHome() {
		if (userHome == null) {
			synchronized(this) {
				userHome = System.getProperty("user.home");
			}
		}
        return userHome;
	}

	protected String getHomeJavaPolicyFilePath() {
		return getUserHome() + '/' + configFileHomePrefix + javaPolicyName;
	}
	protected String getHomeJaasConfigFilePath() {
		return getUserHome() + '/' + configFileHomePrefix + jaasConfigName;
	}
	protected String getDefaultJavaPolicyResourcePath() {
		return configResourcePrefix + javaPolicyName;
	}
	protected String getDefaultJaasConfigResourcePath() {
		return configResourcePrefix + jaasConfigName;
	}
	protected String getDefaultJavaPolicyFilePath() {
		return getUserHome() + '/' + configResourcePrefix + javaPolicyName;
	}
	protected String getDefaultJaasConfigFilePath() {
		return getUserHome() + '/' + configResourcePrefix + jaasConfigName;
	}
	
	/**
	 * Checks Java security (policy, jaas)
	 * and if none sets and outputs a default one
	 */
	public void checkJavaSecurityConfiguration() throws SecurityException {
        if (securityChecked) {
        	// do this method once
        	return;
        }
        
        checkDefaultConfigurationFile(JAVA_POLICY_PROP,
        		getHomeJavaPolicyFilePath(),
        		getDefaultJavaPolicyResourcePath(),
        		getDefaultJavaPolicyFilePath());
        checkDefaultConfigurationFile(JAAS_CONFIG_PROP,
        		getHomeJaasConfigFilePath(),
        		getDefaultJaasConfigResourcePath(),
        		getDefaultJaasConfigFilePath()); 
        
        securityChecked = true; 
	}
	

	protected void checkDefaultConfigurationFile(String propName,
			String homeFilePath, String defaultResourcePath, String defaultFilePath) {
        String propFilePath =  System.getProperty(propName);
        if (propFilePath != null && propFilePath.length() != 0) {
        	// if there is a custom value (indluding ours), do nothing
        	// NB. this allows the user to customize the default policy file
        	return;
        }
        
        if (new File(homeFilePath).exists()) {
        	// If already configured at java default location, let it so
        	// TODO better : handle all default configuration cases, using Sun security util code
        	return;
        }
        
        File defaultFile = new File(defaultFilePath);
        if (defaultFile.exists()) {
	        // if the file already exists but is not set in the property,
	        // then it is NOT overwritten
	    	// NB. this allows the user to customize the default policy file
        	
        } else {
        	// output default configuration
            InputStream defaultResourceIs = this.getClass().getClassLoader()
    				.getResourceAsStream(defaultResourcePath);
            if (defaultResourceIs == null) {
            	throw new SecurityException("Error checking java security : can't find resource "
            			+ defaultResourcePath + " in classpath");
            }
	        try {
	        	if (!defaultFile.createNewFile()) {
	        		// should not happen since does not exist and rights ok since home
	            	throw new SecurityException("Error checking java security : can't create file "
	            			+ defaultFile.getAbsolutePath());
	        	}
				OutputStream defaultPropFileOs
						= new BufferedOutputStream(new FileOutputStream(defaultFile));
				copy(defaultResourceIs, defaultPropFileOs);
			} catch (FileNotFoundException e) {
				// won't happen since created
            	throw new SecurityException("Error checking java security : can't find file "
            			+ defaultFile.getAbsolutePath(), e);
			} catch (IOException e) {
            	throw new SecurityException("Error checking java security : can't copy " +
            			"classpath resource " + defaultResourcePath
            			+ " to file " + defaultFile.getAbsolutePath(), e);
			}
        }
        
        // setting property to default file
        System.setProperty(propName, defaultFilePath);
	}

	protected void copy(InputStream is, OutputStream os) throws IOException {
        byte[] buf = new byte[1024];
        int nbRead;
        while ((nbRead = is.read(buf)) != -1) {
        	os.write(buf, 0, nbRead);
        }
        is.close();
        os.close();
	}
}
