/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id : $ 
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.se.bonita.security;

import java.util.Map;

import javax.security.auth.Subject;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;

import org.objectweb.security.context.SecurityCurrent;

public class PetalsSeBonitaDefaultLoginModule implements LoginModule {

    private Subject subject;
	private CallbackHandler callbackHandler;
    private Map sharedState;
    private Map options;

    // the authentication status
    private boolean succeeded = false;
    private boolean commitSucceeded = false;

    private String username;
    private char[] password;

    private PetalsSeBonitaDefaultPrincipal userPrincipal;

    /**
     * Initialize this <code>LoginModule</code>.
     * 
     * @param subject the <code>Subject</code> to be authenticated.
     * 
     * @param callbackHandler not used
     * 
     * @param sharedState shared <code>LoginModule</code> state.
     * 
     * @param options options specified in the login
     *			<code>Configuration</code> for this particular
     *			<code>LoginModule</code>.
     */
    public void initialize(Subject subject, CallbackHandler callbackHandler,
			Map sharedState, Map options) {
    	this.subject = subject;
    	this.callbackHandler = callbackHandler;
    	this.sharedState = sharedState;
    	this.options = options;
    }

    /**
     * Authenticate the user by prompting for a user name and password.
     *
     * @return true in all cases since this <code>LoginModule</code>
     *		should not be ignored.
     *
     * @exception FailedLoginException if the authentication fails.
     *
     * @exception LoginException if this <code>LoginModule</code>
     *		is unable to perform the authentication.
     */
    public boolean login() throws LoginException {
    	// trivial loginModule
		username = SecurityCurrent.getCurrent().getSecurityContext().getPrincipalName();
		succeeded = true;
    	return succeeded;
    }

    /**
     * <p> This method is called if the LoginContext's
     * overall authentication succeeded
     * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
     * succeeded).</p>
     *
     * <p> If this LoginModule's own authentication attempt
     * succeeded (checked by retrieving the private state saved by the
     * <code>login</code> method), then this method associates a
     * <code>SamplePrincipal</code>
     * with the <code>Subject</code> located in the
     * <code>LoginModule</code>.  If this LoginModule's own
     * authentication attempted failed, then this method removes
     * any state that was originally saved.</p>
     *
     * @exception LoginException if the commit fails.
     *
     * @return true if this LoginModule's own login and commit
     *		attempts succeeded, or false otherwise.
     */
    public boolean commit() throws LoginException {
    	if (!succeeded) {
        	commitSucceeded = false;
    		return commitSucceeded;
    	}
    	
    	// add a Principal (authenticated identity) to the Subject

    	// assume the user we authenticated is the SamplePrincipal
    	userPrincipal = new PetalsSeBonitaDefaultPrincipal(username);
    	if (!subject.getPrincipals().contains(userPrincipal)) {
    		subject.getPrincipals().add(userPrincipal);
    	}

    	// in any case, clean out state
    	username = null;
    	
    	if (password != null) {
    		for (int i = 0; i < password.length; i++) {
    			password[i] = ' ';
    		}
        	password = null;
    	}

    	commitSucceeded = true;
    	return commitSucceeded;
    }

    /**
     * <p> This method is called if the LoginContext's
     * overall authentication failed.
     * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
     * did not succeed).</p>
     *
     * <p> If this LoginModule's own authentication attempt
     * succeeded (checked by retrieving the private state saved by the
     * <code>login</code> and <code>commit</code> methods),
     * then this method cleans up any state that was originally saved.</p>
     *
     * @exception LoginException if the abort fails.
     *
     * @return false if this LoginModule's own login and/or commit attempts
     *		failed, and true otherwise.
     */
    public boolean abort() throws LoginException {
    	if (!succeeded) {
    		return false;
    	}
    	
    	if (succeeded && !commitSucceeded) {
    		// login succeeded but overall authentication failed
    		succeeded = false;
    		username = null;
    		userPrincipal = null;
    		if (password != null) {
    			for (int i = 0; i < password.length; i++) {
    				password[i] = ' ';
    			}
    			password = null;
    		}
    	} else {
    		// overall authentication succeeded and commit succeeded,
    		// but someone else's commit failed
    		logout();
    	}
    	return true;
    }

    /**
     * Logout the user.
     *
     * <p> This method removes the <code>SamplePrincipal</code>
     * that was added by the <code>commit</code> method.</p>
     *
     * @exception LoginException if the logout fails.
     *
     * @return true in all cases since this <code>LoginModule</code>
     *          should not be ignored.
     */
    public boolean logout() throws LoginException {
    	subject.getPrincipals().remove(userPrincipal);
    	succeeded = commitSucceeded;
    	username = null;
    	if (password != null) {
    		for (int i = 0; i < password.length; i++) {
    			password[i] = ' ';
    		}
    		password = null;
    	}
    	userPrincipal = null;
    	return true;
    }
}
